/*
 * template - implements template functionality
 * for LDAP entries.
 *
 * Author: Helmut Schulz <helmut@thalion-graphics.de>
 *
 * (C) 2006,2007 Hewlett-Packard GmbH Germany
 * (C) 2007 Thalion Graphic Solutions eK
 * All Rights Reserved
 *
 * Developed for Hewlett-Packard GmbH Germany
 * Contact: Bernd Schumacher <bernd.schumacher@hp.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted only as authorized by the
 * GNU Public License 2.0.
 *
 * A copy of this license is available at
 * <http://www.fsf.org/licensing/licenses/gpl.txt>
 */

#include "portable.h"

/*#ifdef SLAPD_OVER_TEMPLATE*/

#include <stdio.h>

#include <ac/string.h>
#include <ac/socket.h>

#include "slap.h"
#include "template.h"
#ifdef TEMPLATE_USECACHE
#include "template_cache.h"
#endif


/* int template_is_template    -- returns 1 if AttributeDescription
                              contains the template (0 otherwise)
*/
#define template_isnamed(ad,name) \
 ((ad->ad_cname.bv_len == (name)->bv_len && !strcmp(ad->ad_cname.bv_val,(name)->bv_val)))

typedef struct template_attmod {
	AttributeDescription *m_ad;
	BerVarray             m_vals;
	struct template_attmod *m_next;
#define template_attmod_free(p,op) (ber_bvarray_free_x(p->m_vals,op->o_tmpmemctx),ch_free(p))
} template_attmod_t;

static template_attmod_t *template_attmod_create(AttributeDescription *ad,BerVarray vals,template_attmod_t *next)
{
	template_attmod_t *mod = ch_malloc(sizeof(template_attmod_t));
	mod->m_ad = ad;
	mod->m_vals = vals;
	mod->m_next = next;
	return mod;
}

/*
 * Creates a temporary symbol table, containing all result set attributes.
 * This table is used to determine, if an attribute value needs to be filled in
 * from template defaults.
 * Argument 1: a pointer to the SlapReply structure
 * Returns:    a dynamically allocated template_symtab_t object
 *             (must be deallocated caller-side)
 */
static template_symtab_t *template_create_attmap(SlapReply *rs)
{
	template_symtab_t *symtab = template_symtab_create(TEMPLATE_SMALLHASH,1);
	Attribute *att = rs->sr_entry->e_attrs;
	while ( att ) {
		template_symtab_get(symtab,att->a_desc->ad_cname.bv_val,&att,1);
		att = att->a_next;
	}
	return symtab;
}

/*
 * Look up an attribute description in the result set attribute map:
 * returns true, if the attribute is defined, false otherwise.
 * Argument 1: a pointer to the symbol table
 * Argument 2: a pointer to an AttributeDescription structure
 * Returns: true, if the attribute is defined within the table, false otherwise
 */
static int template_att_isset(template_symtab_t *map,AttributeDescription *ad)
{
	if ( ad ) {
		int sym = template_symtab_get(map,ad->ad_cname.bv_val,NULL,0);
#ifdef TEMPLATE_DEBUG
		fprintf(stderr,"lookup <%s> -> %d\n",ad->ad_cname.bv_val,sym);
#endif
		return sym > -1;
	}
	return 0;
}

#ifdef TEMPLATE_DEBUG
/*
 * For debugging purposes:
 * Callback function for template_symtab_apply() - prints one symtab entry.
 */
static int dbg_printsym(int sym, const char *str, void *info, void *p)
{
	fprintf(stderr,"%s -> %d (info: 0x%X)\n",str,sym,info);
	return 1;
}

/*
 * For debugging purposes:
 * print a symbol table to stderr.
 */
#define dbg_printmap(headline,map) \
 (fprintf(stderr,"%s",(headline)),template_symtab_apply((map),dbg_printsym,NULL))
#else
#define dbg_printmap(headline,map)
#endif

/*
 * The function tests, if the given list of attribute names
 * may be used for template resolution. That is the case, if
 * the list is not empty, and none of the names in it starts
 * with '*' or '@' ('*' should match all defined attributes
 * of the entry, and '@' followed by an object class name
 * should match all defined attributes declared by the
 * object class).
 */
static int template_explicitnames(AttributeName *an)
{
	if ( ! an ) {
		return 0;
	}
	for ( ; an->an_name.bv_val; an++ ) {
		switch(an->an_name.bv_val[0]) {
			case '*':
			case '@':
				return 0;
			default:
				break;
		}
	}
	return 1;
}

/*
 * Response handler.
 * If it is invoked on a search response, it will perform template attribute
 * resolution: for entries containing a value for the template referrer
 * attribute, any attribute value not yet contained in the unmodified result
 * set provided by the backend, will be recursively searched in the referred
 * template entry.
 * For responses on write-requests, the function will issue a cache-invalidate,
 * if caching is enabled.
 */
static int template_response(Operation *op, SlapReply *rs)
{
	slap_overinst *on = (slap_overinst *) op->o_bd->bd_info;
	template_info *hc_main = on->on_bi.bi_private;
	template_ad_info *hc = hc_main->hc_ad_info;
	Attribute *template_att;
	struct berval *template = NULL;
	BerVarray vals = NULL;
	int rc = 0, i;

#ifdef TEMPLATE_DEBUG
	fprintf(stderr,"template_response entered\n");
#endif

	/* If operation failed, no further processing! */
	switch ( rs->sr_err ) {
		case LDAP_SUCCESS:
			/* FIXME: other cases? */
			break;
		default:
			return SLAP_CB_CONTINUE;
	}

#ifdef TEMPLATE_DEBUG
	fprintf(stderr,"template_response: result state==LDAP_SUCCESS\n");
#endif

	/* Only for searches, template resolution must be performed. */
	if (hc && rs->sr_type == REP_SEARCH ) {		
		/* search template */
		template_symtab_t *attmap = template_create_attmap(rs);
		template_attmod_t *mods = NULL;
		AttributeName *an = op->ors_attrs;
		int sym = -1;

		dbg_printmap("template_resonse: result set attributes:\n",attmap);

		/*
		 * First step: check, if the template reference attribute
		 * is defined.
		 */
		if ( (sym = template_symtab_get(attmap,hc_main->templatename.bv_val,&template_att,0)) > -1 ) {
			template = template_att->a_vals;
#ifdef TEMPLATE_DEBUG
			fprintf(stderr,"template <%s->%d>: %s\n",hc_main->templatename.bv_val,sym,template ? template->bv_val : "none");
#endif
		}
		/*
		 * This is partially HP-specific:
		 * the template attribute is undefined, if it's actually not
		 * set, or if it is the string '---'
		 */
		if ( TEMPLATE_ISNIL(hc_main,template) ) {
			template_symtab_free(attmap);
			return SLAP_CB_CONTINUE;
		}
		/*
		 * Next step:
		 * Check, if the search includes an explicit list of attribute
		 * names, and test, if it can be used for template attribute
		 * resolution. This is the case, if no attribute name starts
		 * (or equals - we only test the first char) with '*' or '@'.
		 */
		if ( template_explicitnames(an) ) {
			/*
			 * Okay, we may use the list - as it probably is much
			 * shorter than the list of configured attributes, this
			 * will take much less time.
			 */
#ifdef TEMPLATE_DEBUG
			fprintf(stderr,"template_response: we have an explicit list of names\n");
#endif
			for ( ; an->an_name.bv_val; an++ ) {
				AttributeDescription *ad = an->an_desc;
				if ( ad && (!template_isnamed(ad,&hc_main->templatename)) && (!template_att_isset(attmap,ad)) && (rc = template_resolve_attribute(op,ad,template_att,&vals)) > 0 ) {
					mods = template_attmod_create(ad,vals,mods);
				}
				if ( rc < 0 ) break;
			}
		}
		else {
			/*
			 * No - we must do it the hard way.
			 * */
#ifdef TEMPLATE_DEBUG
			fprintf(stderr,"template_response: have to check all attributes\n");
#endif
			for (hc = hc_main->hc_ad_info,i = 0; hc; hc = hc->hc_next, i++) {
				AttributeDescription *ad = hc->hc_ad;
				if ( (!template_isnamed(ad,&hc_main->templatename)) && (!template_att_isset(attmap,ad)) && (rc = template_resolve_attribute(op,ad,template_att,&vals)) > 0 ) {
					mods = template_attmod_create(ad,vals,mods);
				}
				if ( rc < 0 ) break;
			}	
		}
		/*
		 * If any attributes have been filled in, they must be merged
		 * into the result set.
		 */
		if ( mods ) {
			if ( rc < 0 ) {
				template_attmod_t *modp = mods, *tmp = NULL;
				while ( modp ) {
					tmp = modp->m_next;
					template_attmod_free(modp,op);
					modp = tmp;
				}
			}
			else {
				Entry *new = entry_dup(rs->sr_entry);
				template_attmod_t *modp = mods, *tmp = NULL;
				while ( modp ) {
					tmp = modp->m_next;
					attr_merge(new,modp->m_ad,modp->m_vals,NULL);
					template_attmod_free(modp,op);
					modp = tmp;
				}
				rs->sr_entry = new;
				rs->sr_flags |= REP_ENTRY_MUSTBEFREED;
			}
		}
		template_symtab_free(attmap);
		if ( rc < 0 ) {
			rs->sr_text = "template overlay error";
			rs->sr_err = LDAP_LOOP_DETECT;
			return rc;
		}
		return SLAP_CB_CONTINUE;
	}

#ifdef TEMPLATE_USECACHE
	if ( hc && hc_main->hc_cache ) {
		/* Write operations: cache must be invalidated/updated. */
		switch ( op->o_tag ) {
			case LDAP_REQ_ADD:
			case LDAP_REQ_MODIFY:
			case LDAP_REQ_MODRDN:
			case LDAP_REQ_DELETE:
				break;

			case LDAP_REQ_EXTENDED:
				/* if write, process */
				if ( exop_is_write( op ))
					break;

			/* fall thru */
			default:
				return SLAP_CB_CONTINUE;
		}
		/* 
		 *Okay, cache actually should be invalidated.
		 * At this point, the strategy is simple, based on the assumption,
		 * that write ops will be very rare compared to read events:
		 * erase all cache content.
		 * Performance measurements have to show, if this simple approach
		 * is enough ...
		 */
		template_cache_clear(hc_main->hc_cache);
	}
#endif

	return SLAP_CB_CONTINUE;
}

/*
 * Configuration handler.
 * Sets up overlay configuration from config file.
 * Known config directives currently are:
 * 'template_attributes' - provides the list of template-resolution-enabled
 *                         attributes
 * 'template_base_dn'    - the search base DN for template resolution
 * 'template_cachelines' - an integer value specifying the maximum number of
 *                         cacheable template UIDs
 * 'template_name'       - the name of the template referrer attribute
 *                         (available only, if compiled with TEMPLATE_USECACHE
 *                         macro enabled)
 * 'template_nilvalue'   - the NIL value string for undefined attributes
 *                         (available only, if built with HP_NILVALUE macro
 *                         enabled)
 */
static int template_db_config(BackendDB	*be,const char *fname,
                            int lineno,int argc,char **argv)
{
	int i;
	const char *text;
	slap_overinst *on = (slap_overinst *) be->bd_info;
	template_info *hc;
	Debug( LDAP_DEBUG_CONFIG,"template_db_config enter\n",0,0,0);
	if (!on->on_bi.bi_private) {
		hc = ch_malloc(sizeof(template_info));
		memset(hc,0,sizeof(template_info));
		if ( ! template_symtab_init(&hc->symtab,TEMPLATE_LARGEHASH,0) ) {
			Debug(LDAP_DEBUG_ANY,
				"%s: line %d: could not initialize attribute symbol table\n",0,0,0);
			return 1;
		}
#ifdef HP_NILVALUE
		hc->nilvalue = HP_NILVALUE;
#endif
#ifdef TEMPLATE_USECACHE
		hc->cachelines = 100;
#endif
		on->on_bi.bi_private = hc;
	}
	else
		hc = on->on_bi.bi_private;
	if (strcasecmp(argv[0],"template_attributes") == 0 ) {
		for (i = 1; i < argc; i++) {
			AttributeDescription *ad = NULL;
			if (slap_str2ad(argv[i],&ad,&text)) {
				Debug(LDAP_DEBUG_ANY,
					"%s: line %d: attribute description unknown in \"template_attributes\" line: %s.\n",
					fname,lineno,text);
				return 1;
			}
			else {
				Debug(LDAP_DEBUG_CONFIG,"%s: line %d: handle attribute %s.\n",
					fname, lineno, argv[i]);
				template_ad_info *hc_ad;
				hc_ad = ch_malloc(sizeof(template_ad_info));
				hc_ad->hc_ad = ad;
				hc_ad->hc_next = hc->hc_ad_info;
				hc->hc_ad_info = hc_ad;
				hc->attribute_cnt++;
				if ( ! template_symtab_get(&hc->symtab,argv[i],NULL,1) ) {
					Debug(LDAP_DEBUG_CONFIG,"%s: line %d: could not add %s to symbol table.\n",
						fname, lineno, argv[i]);
				}
			}
		}
		return 0;
	}
#ifdef TEMPLATE_USECACHE
	if (strcasecmp(argv[0],"template_cachelines") == 0 ) {
		if ( argc != 2 ) {
			Debug(LDAP_DEBUG_ANY,
				"%s: line %d: a single non-negative integer is expected here\n",
				fname,lineno,0);
			return 1;
		}
		hc->cachelines = atoi(argv[1]);
		if ( hc->cachelines < 0 ) {
			Debug(LDAP_DEBUG_ANY,
				"%s: line %d: a single non-negative integer is expected here\n",
				fname,lineno,0);
			return 1;
		}
		return 0;
	}
#endif
	if ( strcasecmp(argv[0],"template_name") == 0 ) {
		if (argc < 2) {
			Debug(LDAP_DEBUG_ANY,
				"%s: line %d: template attribute name missing in \"template_name\"\n",
				fname,lineno,0);
			return 1;
		}
		ber_str2bv(argv[1],0,1,&hc->templatename);
		return 0;
	}
#ifdef HP_NILVALUE
	if (strcasecmp(argv[0],"template_nilvalue") == 0 ) {
		if (argc < 2) {
			Debug(LDAP_DEBUG_ANY,
				"%s: line %d: nilvalue missing in \"template_nilvalue\"\n",
				fname,lineno,0);
			return 1;
		}
		hc->nilvalue = ch_strdup(argv[1]);
		return 0;
	}
#endif
	if (strcasecmp(argv[0],"template_base_dn") == 0 ) {
		if (argc < 2) {
			Debug(LDAP_DEBUG_ANY,
				"%s: line %d: base dn missing in \"template_base_dn\"\n",
				fname,lineno,0);
			return 1;
		}
		hc->base_dn = ch_strdup(argv[1]);
		return 0;
	}
	return SLAP_CONF_UNKNOWN;
}

#ifdef TEMPLATE_USECACHE
/*
 * Open handler.
 * Allocates cache.
 */
static int template_db_open(BackendDB *be)
{
	slap_overinst *on = (slap_overinst *) be->bd_info;
	template_info *hc = (template_info *) on->on_bi.bi_private;
	dbg_printmap("template: configured attributes:\n",&hc->symtab);
	if ( hc->cachelines ) {
		hc->hc_cache = template_cache_create(hc->attribute_cnt,hc->cachelines);
		if ( ! hc->hc_cache ) {
			Debug(LDAP_DEBUG_ANY,"template_db_open: failed to setup attribute cache\n",0,0,0);
			return 1;
		}
	}
	else {
		hc->hc_cache = NULL;
		Debug(LDAP_DEBUG_ANY,"template_db_open: cache deactivated\n",0,0,0);
	}
	return 0;
}
#endif

/*
 * DB close handler.
 * Deallocates symbol table and cache.
 */
static int
template_db_close(BackendDB *be)
{
	slap_overinst *on = (slap_overinst *)be->bd_info;
	template_info *hc = (template_info *) on->on_bi.bi_private;
	template_symtab_deinit(&hc->symtab);
#ifdef TEMPLATE_USECACHE
	if ( hc->hc_cache ) {
		template_cache_free(hc->hc_cache);
	}
#endif
	ch_free(hc);
	return 0;
}

static slap_overinst template;

/* init routine */
int template_initialize() 
{
	Debug( LDAP_DEBUG_CONFIG,"template_initialize enter\n",0,0,0);
	template.on_bi.bi_type = "template";
	template.on_response = template_response;
	template.on_bi.bi_db_config = template_db_config;
#ifdef TEMPLATE_USECACHE
	template.on_bi.bi_db_open = template_db_open;
#endif
	template.on_bi.bi_db_close = template_db_close;
	return overlay_register(&template);
}

/*#if SLAPD_OVER_TEMPLATE == SLAPD_MOD_DYNAMIC*/
#if 1
int init_module(int argc, char *argv[]) {
	return template_initialize();
}
#endif

/*#endif /* SLAPD_OVER_COLLECT */
