/*
 * template_cache.c
 * Provides a attribute value cache for template entries.
 * The cache is implemented as sparsely populated matrix of dimension
 * <number of schema attributes> X <maximum number of cacheable entries>.
 *
 * Cache indices are computed by looking up attribute names and entry
 * identifiers (i.e. 'cn' or 'uid' attribute values) in symbol tables,
 * which keep unique integer values for each registered string.
 *
 * Thus, access to saved values in constant time is granted.
 *
 * Author: H. Schulz <helmut@thalion-graphics.de>
 * (C) Thalion Graphic Solutions eK. 2007, all rights reserved
 *
 * Developed for Hewlett-Packard GmbH Germany
 * Contact: Bernd Schumacher <bernd.schumacher@hp.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted only as authorized by the
 * GNU Public License 2.0.
 *
 * A copy of this license is available at
 * <http://www.fsf.org/licensing/licenses/gpl.txt>
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "template_cache.h"
#include "template_symtab.h"

#include "template_malloc.h"

#ifdef TEMPLATE_CACHE_STANDALONETEST
#include <stdio.h>
#include <unistd.h>

#define NATTR_DEFAULT 1000
#define NCOS_DEFAULT 5
#define NITER_DEFAULT 1

template_cachedval_t copyval_x(template_cachedval_t v,void *p)
{
	return (template_cachedval_t) strdup((char *)v);
}
template_cachedval_t copyval(template_cachedval_t v)
{
	return (template_cachedval_t) strdup((char *)v);
}

void freeval(template_cachedval_t v)
{
	free(v);
}

#else

/*
 * ** dup_bervarray()
 * **      copy a BerVarray;
 * */

static BerVarray dup_bervarray_x(BerVarray b,void *ctx)
{
	int i, len;
	BerVarray nb;
	for(len = 0; b[len].bv_val; len++);
	nb = ch_malloc((len+1) * sizeof(BerValue));
        for(i = 0; i < len; i++) ber_dupbv(&nb[i], &b[i]);
	nb[len].bv_val = NULL;
	nb[len].bv_len = 0;
	return(nb);
}


template_cachedval_t copyval_x(template_cachedval_t v,void *ctx)
{
	return dup_bervarray_x((BerVarray)v,ctx);
}

template_cachedval_t copyval(template_cachedval_t v)
{
	return dup_bervarray_x((BerVarray)v,NULL);
}

void freeval(template_cachedval_t v)
{
	ber_bvarray_free(v);
}

#endif

/*
 * The cache area is organized as sparsely populated matrix [nattr x ncos]
 */
struct template_cache { 
	template_symtab_t m_costab;
	template_cachedval_t *m_values;
	int m_nattr;
	int m_ncos;
#ifndef TEMPLATE_CACHE_STANDALONETEST
	ldap_pvt_thread_mutex_t	m_mutex;
#endif
	template_cachedval_t (*m_copyval)(template_cachedval_t);
	template_cachedval_t (*m_copyval_x)(template_cachedval_t,void *);
	void (*m_freeval)(template_cachedval_t);
#define cache_cossym(cache,str) (template_symtab_get(&cache->m_costab,(str),NULL,1))
};

template_cache_t *template_cache_create(int nattr, int ncos)
{
	template_cache_t *cache = TEMPLATE_MALLOC(sizeof(template_cache_t));
	if ( ! template_cache_init(cache,nattr,ncos) ) {
		TEMPLATE_FREE(cache);
		return NULL;
	}
	return cache;
}

template_cache_t *template_cache_init(template_cache_t *cache, int nattr, int ncos)
{
	int cachesize = sizeof(template_cachedval_t) * (nattr * ncos);
	if ( ! template_symtab_init(&cache->m_costab,TEMPLATE_LARGEHASH,0) ) return NULL;
	if ( ! (cache->m_values = TEMPLATE_MALLOC(cachesize)) ) return NULL;
	memset(cache->m_values,0,cachesize); 
	cache->m_nattr = nattr;
	cache->m_ncos = ncos;
	cache->m_copyval_x = copyval_x;
	cache->m_copyval = copyval;
	cache->m_freeval = freeval;
#ifndef TEMPLATE_CACHE_STANDALONETEST
	ldap_pvt_thread_mutex_init(&cache->m_mutex);
#endif
#ifdef TEMPLATE_DEBUG
	fprintf(stderr,"cache matrix: %d x %d, size %d\n",nattr,ncos,cachesize);
#endif
	return cache;
}

template_cachedval_t template_cache_get(template_cache_t *cache,int attsym,const char *cos,int copy,void *ctx)
{
	int cossym;
	template_cachedval_t rv = NULL;
#ifdef TEMPLATE_DEBUG
	fprintf(stderr,"att-id <%d>, cos <%s>\n", attsym, cos); 
#endif
#ifndef TEMPLATE_CACHE_STANDALONETEST
	ldap_pvt_thread_mutex_lock( &cache->m_mutex );
#endif
	cossym = cache_cossym(cache,cos);
#ifdef TEMPLATE_DEBUG
	fprintf(stderr,"cos <%s> -> id <%d>\n", cos, cossym); 
#endif
	if ( attsym < cache->m_nattr && cossym < cache->m_ncos ) {
		rv = cache->m_values[(cache->m_nattr * cossym) + attsym];
#ifdef TEMPLATE_DEBUG
		fprintf(stderr,"returning val <0x%X>\n",rv);
#endif
		if ( rv && copy ) {
			rv = cache->m_copyval_x(rv,ctx);
		}
	}
#ifndef TEMPLATE_CACHE_STANDALONETEST
	ldap_pvt_thread_mutex_unlock( &cache->m_mutex );
#endif
	return rv;
}

int template_cache_put(template_cache_t *cache,int attsym,const char *cos,template_cachedval_t val)
{
	int cossym, rv = 0;
#ifdef TEMPLATE_DEBUG
	fprintf(stderr,"att-id <%d>, cos <%s>\n", attsym, cos); 
#endif
#ifndef TEMPLATE_CACHE_STANDALONETEST
	ldap_pvt_thread_mutex_lock( &cache->m_mutex );
#endif
	cossym = cache_cossym(cache,cos);
#ifdef TEMPLATE_DEBUG
	fprintf(stderr,"cos <%s> -> id <%d>\n", cos, cossym); 
#endif
	if ( attsym < cache->m_nattr && cossym < cache->m_ncos ) {
#ifdef TEMPLATE_DEBUG
		fprintf(stderr,"saving val <0x%X> @ (%d,%d)\n",val,attsym,cossym);
#endif
		cache->m_values[cache->m_nattr * cossym + attsym] = cache->m_copyval(val);
		rv = 1;
	}
#ifndef TEMPLATE_CACHE_STANDALONETEST
	ldap_pvt_thread_mutex_unlock( &cache->m_mutex );
#endif
	return rv;
}

void template_cache_clear(template_cache_t *cache)
{
	int i = 0;
	int n = cache->m_nattr * cache->m_ncos;
#ifndef TEMPLATE_CACHE_STANDALONETEST
	ldap_pvt_thread_mutex_lock( &cache->m_mutex );
#endif
	for ( ; i<n; i++ ) {
		cache->m_freeval(cache->m_values[i]);
		cache->m_values[i] = NULL;
	}
#ifndef TEMPLATE_CACHE_STANDALONETEST
	ldap_pvt_thread_mutex_unlock( &cache->m_mutex );
#endif
}

void template_cache_deinit(template_cache_t *cache)
{
	int i = 0;
	int n = cache->m_nattr * cache->m_ncos;
	template_symtab_deinit(&cache->m_costab);
	for ( ; i<n; i++ ) {
		cache->m_freeval(cache->m_values[i]);
	}
	TEMPLATE_FREE(cache->m_values);
#ifndef TEMPLATE_CACHE_STANDALONETEST
	ldap_pvt_thread_mutex_destroy(&cache->m_mutex);
#endif
	
}

void template_cache_free(template_cache_t *cache)
{
	template_cache_deinit(cache);
	TEMPLATE_FREE(cache);
}

#ifdef TEMPLATE_CACHE_STANDALONETEST

template_cache_t *cache = NULL;
template_symtab_t *symtab = NULL;

void help()
{
	printf("This utility executes a hp_cos cache standalone test.\n");
	printf("A set of plain ascii text files is read, and each text line found is used to\n");
	printf("setup a  cache  entry associated  to the tuple  (line<lineno> x <filename>),\n");
	printf("representing the (attribute x cos)  tuple  within  the  real-world scenario.\n");
	printf("After the  setup  part  is done,  the cache  is sequentially queried for all\n");
	printf("tuples  (line<i> x <filename>) with i>=0 and  i<number  of attributes in the\n");
	printf("cache.\n");
	printf("When completed, the cache structure is freed.\n");
	printf("Optionally, the test can be executed repeatedly, thus exercising cache setup\n");
	printf("usage and cleanup.\n");
}

void usage(char *name)
{
	printf("usage: %s [-a][-c][-n][-h]\n",name);
	printf("-a <nattr> set the maximum number of cacheable attributes (default: %d)\n",NATTR_DEFAULT);
	printf("-c <ncos>  set the maximum number of cacheable service classes (default: %d)\n",NCOS_DEFAULT);
	printf("-n <niter> set the number of cycles the test should run (default: %d)\n",NITER_DEFAULT);
	printf("-h         print this text\n");
}

#define DELIM " \t\n"

int suckkeys(const char *name)
{
	FILE *f = fopen(name,"r");
	int line = 0;
	if ( ! f ) {
		fprintf(stderr,"open '%s' failed.\n",name);
		return 0;
	}
	while ( ! feof(f) ) {
		char line_str[1000];
		if ( fgets(line_str,1000,f) ) {
				char *key = strtok(line_str,DELIM);
				for ( ; key; key = strtok(NULL,DELIM) ) {
					int sym = template_symtab_get(symtab,key,NULL,1);
					printf("%s -> %d\n",key,sym);
				}
		}
		line++;
	}
	fclose(f);
	return 1;
}

int lookup(int sym, const char *str, void *info, void *p)
{
	char *v = template_cache_get(cache,sym,(char *)p,0,NULL);
	if ( v ) {
		printf("%s (symbol %d) occurs in %s: %s\n",str,sym,(char *)p,v);
	}
	return 1;
}

int runtest(int optind,int argc,char **argv)
{
	int i = optind;
	for ( ; i < argc; i++ ) {
		FILE *f = fopen(argv[i],"r");
		int line = 0;
		if ( ! f ) {
			fprintf(stderr,"open '%s' failed.\n",argv[i]);
			return 0;
		}
		while ( ! feof(f) ) {
			char line_str[1000];
			if ( fgets(line_str,1000,f) ) {
				printf("line: %s\n",line_str);
				char *token = strtok(line_str,DELIM);
				for ( ; token; token = strtok(NULL,DELIM) ) {
					int sym = template_symtab_get(symtab,token,NULL,0);
					printf("%s -> %d\n",token,sym);
					if ( sym >= 0 ) {
						char *rv = template_cache_get(cache,sym,argv[i],0,NULL);
						if ( ! rv ) {
							fputc( template_cache_put(cache,sym,argv[i],line_str) ? '+' : '0',stdout);
						}
						else {
							printf("already cached: %s\n",rv);
						}
					}
				}
			}
			line++;
		}
		fclose(f);
	}
	for ( i=optind; i < argc; i++ ) {
		printf("symbols in <%s>:\n",argv[i]);
		template_symtab_apply(symtab,lookup,argv[i]);
	}
	return 1;
}

int main(int argc,char *argv[])
{
	int c;
	extern char *optarg;
	extern int optind;
	int nattr = NATTR_DEFAULT;
	int ncos = NCOS_DEFAULT;
	int n = NITER_DEFAULT;
	char *progname = argv[0];
	char *keyfilename = "keys.txt";
	while ( (c = getopt(argc, argv, "a:c:k:n:h")) != -1 ) {
		switch ( c ) {
			case 'a':
				nattr = atoi(optarg);
				break;
			case 'c':
				ncos = atoi(optarg);
				break;
			case 'k':
				keyfilename = optarg;
				break;
			case 'n':
				n = atoi(optarg);
				break;
			case 'h':
				usage(progname);
				help();
				return 0;
			default:
				fprintf(stderr,"invalid cmd option -%c\n",c);
				usage(progname);
				return -1;
		}
	}
	if ( optind == -1 || optind >= argc ) {
		return 0;
	}
	else {
		int i = 0;
		for ( ; i<n; i++ ) {
			symtab = template_symtab_create(TEMPLATE_LARGEHASH,0);
			if ( ! symtab ) {
				fprintf(stderr,"symtab creation failed\n",i);
				return -1;
			}
			if ( ! suckkeys(keyfilename) ) {
				fprintf(stderr,"symtab setup from key file failed\n",i);
				return -1;
			}
			cache = template_cache_create(symtab->m_lastsym,ncos);
			if ( ! cache ) {
				fprintf(stderr,"cache creation failed\n",i);
				return -1;
			}
			if ( ! runtest(optind,argc,argv) ) {
				fprintf(stderr,"test #%d failed\n",i);
				return -1;
			}
			template_cache_free(cache);
			template_symtab_free(symtab);
			cache = NULL;
			symtab = NULL;
		}
	}
	return 0;
}

#endif
