/*
 * template_resolve.c
 * Implements the attribute resolution algorithm used by the template
 * overlay:
 * + check, if attribute x template-ID is cached:
 *   + yes: return cached value
 *   + no:
 *     + for ever:
 *       + check, if template ID has already been visited:
 *         + yes: return ERROR
 *         + no:
 *           + query backend for attribute in template entry
 *           + if result set not empty:
 *             + cache result set for original attribute x template ID 
 *             + return result set
 *           + else:
 *             + query backend for template ID in template entry
 *             + if result set is valid template ID:
 *               template-ID = result set
 *             + else:
 *               + cache NIL for orignal attribute x template ID
 *               + return NIL
 *
 * Author: H. Schulz <Helmut@thalion-graphics.de>
 *
 * Developed for Hewlett-Packard GmbH Germany
 * Contact: Bernd Schumacher <bernd.schumacher@hp.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted only as authorized by the
 * GNU Public License 2.0.
 *
 * A copy of this license is available at
 * <http://www.fsf.org/licensing/licenses/gpl.txt>
 */
#include "portable.h"
#include "slap.h"
#include "template.h"

#define TEMPLATE_STRBUFSIZ 1000

typedef struct templateentry {
	char            *m_template;
	struct templateentry *m_next;
} templateentry_t;

static templateentry_t *templateentry_create(char *template,templateentry_t *next)
{
	templateentry_t *entry = (templateentry_t *) ch_malloc(sizeof(templateentry_t));
	entry->m_template = template;
	entry->m_next = next;
	return entry;
}

static int template_visited(templateentry_t *list, const char *template)
{
	for ( ; list; list = list->m_next ) {
		if ( ! strcmp(template,list->m_template) ) {
			return 1;
		}
	}
	return 0;
}

#define templateentry_free(entry) (ch_free((entry)->m_template),ch_free(entry))

/* If TEMPLATE_USECACHE is set, this code needs proper locking!!!! TBD. */
int template_resolve_attribute(Operation *op, AttributeDescription *ad, Attribute *template_att, BerVarray *vals)
{
	char tmpl_dn_str[TEMPLATE_STRBUFSIZ];
	BerVarray tmpvals = NULL;
	struct berval tmpl_dn = {NULL,0};
	struct berval bv = {NULL,0};
	char *template_str = ch_strdup(template_att->a_vals->bv_val);
	AttributeDescription *template_ad = template_att->a_desc;
	slap_overinst *on = (slap_overinst *) op->o_bd->bd_info;
	template_info *hc_main = on->on_bi.bi_private;
	templateentry_t *tmpllist = NULL;
	int rv = 0;
	int rc = 0;
#ifdef TEMPLATE_DEBUG
	fprintf(stderr,"template_resolve_attribute(%s,%s)\n",ad->ad_cname.bv_val,template_str);
#endif
#ifdef TEMPLATE_USECACHE
	if ( hc_main->hc_cache ) {
		int attsym = template_symtab_get(&hc_main->symtab,ad->ad_cname.bv_val,NULL,0);
		if ( attsym >= 0 )
			*vals = template_cache_get(hc_main->hc_cache,attsym,template_str,1,op->o_tmpmemctx);
		else
			*vals = NULL;
		if ( *vals ) {
#ifdef TEMPLATE_DEBUG
			fprintf(stderr,"returning cached values\n");
#endif
			ch_free(template_str);
			return 1;
		}
	}
#endif
	for (;;) {
#ifdef TEMPLATE_DEBUG
		fprintf(stderr,"not cached - doing full lookup\n");
#endif
		if ( template_visited(tmpllist,template_str) ) {
			Debug(LDAP_DEBUG_ANY,"loop in COS dependencies detected",0,0,0);
			rv = -1;
			break;
			
		}
		/* create template_dn */
		if ( TEMPLATE_STRBUFSIZ <= strlen(template_str) + strlen(hc_main->base_dn) + 6 ) {
			Debug(LDAP_DEBUG_ANY,"tmpl DN string exceeds max buffer size",0,0,0);
			rv = -1;
			break;
		}
		sprintf(tmpl_dn_str,"uid=%s,%s",template_str,hc_main->base_dn);
#ifdef TEMPLATE_DEBUG
		fprintf(stderr,"NEW DN:%s -------------------------------------\n",tmpl_dn_str);
#endif
		ber_str2bv(tmpl_dn_str,0,0,&bv);
		if (dnNormalize(0, NULL, NULL, &bv, &tmpl_dn, NULL)) {
			Debug(LDAP_DEBUG_ANY,"invalid DN in template_response TODO Dn",0,0,0);
			rv = -1;
			break;
        	}
		*vals = NULL;
	        rc = backend_attribute(op, NULL, &tmpl_dn, ad, vals, ACL_READ);
        	if ( TEMPLATE_ISNIL(hc_main,*vals) ) {
			/*
			 * Need this, because TEMPLATE_ISNIL() may check
			 * for other conditions as well ...
			 */
			if ( *vals )
				ber_bvarray_free_x(*vals,op->o_tmpmemctx);
			*vals = NULL;
		}
		else {
			/* GOTCHA - return success! */
#ifdef TEMPLATE_USECACHE
			if ( hc_main->hc_cache ) {
				/* Use the toplevel template as key. */
				char *template_str = template_att->a_vals->bv_val;
				/* The value must be cached as a duplicate in global
				 * memory context! */
				int attsym = template_symtab_get(&hc_main->symtab,ad->ad_cname.bv_val,NULL,0);
				if ( attsym > -1 )
					template_cache_put(hc_main->hc_cache,attsym,template_str,*vals);
			}
#endif
 			rv = 1;
			break;
		}

		tmpvals = NULL;
		rc = backend_attribute(op, NULL, &tmpl_dn, template_ad, &tmpvals, ACL_READ);
		if ( TEMPLATE_ISNIL(hc_main,tmpvals) ) {
			/*
			 * template not set in COS - return.
			 * For line below see comments concerning 'vals'  above.
			 */
			if ( tmpvals )
				ber_bvarray_free_x(tmpvals,op->o_tmpmemctx);
			rv = 0;
			break;
		}
		tmpllist = templateentry_create(template_str,tmpllist);
		ch_free(tmpl_dn.bv_val);
		tmpl_dn.bv_len = 0;
		tmpl_dn.bv_val = NULL;
		template_str = ch_strdup(tmpvals[0].bv_val);
		ber_bvarray_free_x(tmpvals,op->o_tmpmemctx);
		/* Start new iteration with new template - new game, new luck. */
	}
#ifdef TEMPLATE_USECACHE
	if ( hc_main->hc_cache && !rv ) {
		/* Use the toplevel template as key. */
		char *template_str = template_att->a_vals->bv_val;
		/* Non-set attributes must be cached as such as well.
		 * Otherwise we loose most of the speed gain.  */
		struct berval nil = { NULL, 0 };
		int attsym = template_symtab_get(&hc_main->symtab,ad->ad_cname.bv_val,NULL,0);
		if ( attsym > -1 )
			template_cache_put(hc_main->hc_cache,attsym,template_str,&nil);
	}
#endif
	if ( tmpl_dn.bv_val )
		ch_free(tmpl_dn.bv_val);
	if ( template_str ) 
		ch_free(template_str);
	while ( tmpllist ) {
		templateentry_t *tmp = tmpllist->m_next;
		templateentry_free(tmpllist);
		tmpllist = tmp;
	}
	return rv;
}

