/*
 * template_symtab.c
 * A symbol table for the template overlay.
 * The symbol table associates strings with unique integer
 * identifiers, and can be used to store optional information
 * specific for this string.
 *
 * It is used to store configured attributes, result set attributes,
 * and cached template identifiers.
 *
 * Author: H. Schulz <helmut@thalion-graphics.de>
 * (C) Thalion Graphic Solutions eK. 2007, all rights reserved.
 * 
 * Developed for Hewlett-Packard GmbH Germany
 * Contact: Bernd Schumacher <bernd.schumacher@hp.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted only as authorized by the
 * GNU Public License 2.0.
 *
 * A copy of this license is available at
 * <http://www.fsf.org/licensing/licenses/gpl.txt>
 */
#include <stdlib.h>
#include <string.h>
#ifdef TEMPLATE_DEBUG
#include <stdio.h>
#include <assert.h>
#endif

#include "template_symtab.h"
#include "template_malloc.h"

struct symrec {
	char *m_str;
	int m_id;
	void *m_info;
	struct symrec *m_next_inbucket;
	struct symrec *m_next;
#define symrec_str(symrec) ((symrec)->m_str)
#define symrec_id(symrec) ((symrec)->m_id)
#define symrec_next_inbucket(symrec) ((symrec)->m_next_inbucket)
#define symrec_next(symrec) ((symrec)->m_next)
#define symrec_info(symrec) ((symrec)->m_info)
#define symrec_init(symrec,str,id,next,next_inbucket) \
 ((symrec)->m_str = TEMPLATE_STRDUP(str),(symrec)->m_id = (id),(symrec)->m_next = (next),(symrec)->m_next_inbucket = (next_inbucket),(symrec))
#define symrec_init_shallow(symrec,str,id,next,next_inbucket) \
 ((symrec)->m_str = (str),(symrec)->m_id = (id),(symrec)->m_next = (next),(symrec)->m_next_inbucket = (next_inbucket),(symrec))
#define symrec_free(rec) (TEMPLATE_FREE((rec)->m_str),TEMPLATE_FREE(rec))
};

static int strhash(const char *s)
{
	int l = s ? strlen(s) : 0;
	return l ? (s[0] << 24 | s[l/2] << 16 | s[l/2+1] << 8 | s[l-1]) : 0;
}

template_symrec_t *symrec_new(const char *str,int id,void *info,template_symrec_t **root,template_symrec_t *next_inbucket,int shallow)
{
	template_symrec_t *symrec = (template_symrec_t *) TEMPLATE_MALLOC(sizeof(template_symrec_t));
	symrec->m_str = shallow ? str : TEMPLATE_STRDUP(str);
	symrec->m_id = id;
	symrec->m_info = info;
	symrec->m_next = *root;
	symrec->m_next_inbucket = next_inbucket;
	*root = symrec;
	return symrec;
}

template_symtab_t *template_symtab_create(int nhashbuckets,int shallow)
{
	template_symtab_t *symtab = (template_symtab_t *) TEMPLATE_MALLOC(sizeof(template_symtab_t));
	if ( symtab ) {
		if ( template_symtab_init(symtab,nhashbuckets,shallow) ) {
			return symtab;
		}
		else {
			TEMPLATE_FREE(symtab);
		}
	}
	return NULL;	
}

template_symtab_t *template_symtab_init(template_symtab_t *symtab,int nhashbuckets,int shallow)
{
	size_t size = sizeof(template_symrec_t *) * nhashbuckets;
	symtab->m_buckets = (template_symrec_t **) TEMPLATE_MALLOC(size);
	memset(symtab->m_buckets,0,size);
	symtab->m_nhashbuckets = nhashbuckets;
	symtab->m_root = NULL;
	symtab->m_func = strhash;
	symtab->m_lastsym = 0;
	symtab->m_shallow = shallow;
	return symtab;
}

int template_symtab_get(template_symtab_t *symtab,const char *str,void **info,int create)
{
	int idx = symtab->m_func(str) % symtab->m_nhashbuckets;
	template_symrec_t **bucket = &symtab->m_buckets[idx];
	template_symrec_t *rec = *bucket;
	for ( ; rec; rec = symrec_next_inbucket(rec) ) {
		if ( ! strcmp(str,symrec_str(rec)) ) {
#ifdef TEMPLATE_DEBUG
			fprintf(stderr,"symtab match found: %s->%d\n",str,symrec_id(rec));
#endif
			if ( info ) {
#ifdef TEMPLATE_DEBUG
				fprintf(stderr,"info = 0x%X - filling in saved info 0x%X\n",info,symrec_info(rec));
#endif
				*info = symrec_info(rec);
			}
			return symrec_id(rec);
		}
	}
	/* No match found - add new entry. */
	if ( ! create ) {
		return -1;
	}
	rec = symrec_new(str,symtab->m_lastsym++,info?*info:NULL,&symtab->m_root,*bucket,symtab->m_shallow);
	*bucket = rec;
	/*printf("symtab new entry: %s->%d\n",str,symrec_id(rec));*/
	return symrec_id(rec);
}

int template_symtab_apply(template_symtab_t *symtab,int (*f)(int sym, const char *str, void *info, void *data),void *data)
{
	template_symrec_t *rec = symtab->m_root;
	for ( ; rec; rec = symrec_next(rec) ) {
		if ( ! (*f)(symrec_id(rec),symrec_str(rec),symrec_info(rec),data) ) {
			return 0;
		}
	}
	return 1;
}

void template_symtab_deinit(template_symtab_t *symtab)
{
	int i=0;
	template_symrec_t *rec = symtab->m_root;
	while ( rec ) {
		template_symrec_t *next = symrec_next(rec);
		if ( symtab->m_shallow ) {
			TEMPLATE_FREE(rec);
		}
		else {
			symrec_free(rec);
		}
		rec = next;
	}
	TEMPLATE_FREE(symtab->m_buckets);
	symtab->m_buckets = NULL;
	symtab->m_nhashbuckets = 0;
	symtab->m_root = NULL;
	symtab->m_lastsym = 0;
}

void template_symtab_free(template_symtab_t *symtab)
{
	template_symtab_deinit(symtab);
	TEMPLATE_FREE(symtab);
}

#ifdef TEMPLATE_SYMTAB_STANDALONETEST

#include <stdio.h>
#include <pthread.h>

char *words[] = { "foo", "bar", "baz", "foobar", "foobaz", "fizzle", "drofnats" };
int nwords = 7;
int nthreads = 5;
int niter = 10000000;

static void runtest(void *data)
{
	int i = 0;
	int id = (int *)data;
	for ( ; i<niter; i++ ) {
		template_symtab_t *symtab = template_symtab_create(TEMPLATE_SMALLHASH,0);
		int j = 0;
		for ( ; j<nwords; j++ ) {
			int sym = template_symtab_get(symtab,words[j],NULL,1);
		}
		for ( ; j<nwords; j++ ) {
			if ( j != template_symtab_get(symtab,words[j],NULL,0) ) {
				fprintf(stderr,"symbol <%s> not matched\n",words[j]);
				return;
			}
		}
		template_symtab_free(symtab);
		/*printf("<%d:%d>\n",id,i);fflush(stdout);*/
	}
}

int main(int argc, char *argv[])
{

	int i = 0;
	pthread_t *tid = (pthread_t *) malloc(sizeof(pthread_t) * nthreads);
	for ( ; i<nthreads; i++ ) {
		if ( pthread_create(&tid[i],NULL,&runtest,(void *)i) ) {
			fprintf(stderr,"failed to create thread <%d>\n",i);
			return -1;
		}
	}
	for ( i=0; i<nthreads; i++ ) {
		pthread_join(tid[i],NULL);
	}
	return 0;
}

#endif
