#ifndef TEMPLATE_SYMTAB_H
#define TEMPLATE_SYMTAB_H

/*
 * A symbol table used within template overlay functions.
 *
 * Author: Helmut Schulz <helmut@thalion-graphics.de>
 *
 * Developed for Hewlett-Packard GmbH Germany
 * Contact: Bernd Schumacher <bernd.schumacher@hp.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted only as authorized by the
 * GNU Public License 2.0.
 *
 * A copy of this license is available at
 * <http://www.fsf.org/licensing/licenses/gpl.txt>
 */

typedef struct symrec template_symrec_t; 

/*
 * template_symtab is a rather dumb, non-self-adapting hash map, ie.
 * it keeps the number of hash buckets it has been initialized with.
 * For this reason, it is essential to select a proper bucket count
 * when instantiating template_symtab.
 * Below, two suitable sizes are predefined: the small one for
 * temporaries used to buffer result set attribute names, the big
 * one for the large, static symbol table which maps configured
 * attributes to symbol IDs.
 */

#define TEMPLATE_SMALLHASH 67
#define TEMPLATE_LARGEHASH 1009

typedef struct {
	template_symrec_t **m_buckets;
	template_symrec_t *m_root;
	int m_nhashbuckets;
	int (*m_func)(const char *);
	int m_shallow;
	int m_lastsym;
} template_symtab_t;

/*
 * Create a symbol table instance.
 * Argument 1: the number of hash buckets to use
 * Argument 2: flag indicating whether the symbol table should make copies
 *             of saved strings or not - 0 = yes, all other = no
 * Returns:    a new instance of template_symtab or
 *             NULL, if the function fails
 */
template_symtab_t *template_symtab_create(int nhashbuckets,int shallow);
/*
 * Initializes a symbol table instance in caller-provided memory.
 * Argument 1: a pointer to a memory location of at least sizeof(template_symtab_t) bytes
 * Argument 2: the number of hash buckets to use
 * Argument 3: flag indicating whether the symbol table should make copies
 *             of saved strings or not - 0 = yes, all other = no
 * Returns:    the pointer to the now set-up template_symtab instance, or NULL, if setup
 *             failed
 */
template_symtab_t *template_symtab_init(template_symtab_t *symtab,int nhashbuckets,int shallow);
/*
 * Returns the symbol ID of a given string from this symbol table
 * Argument 1: a pointer to the template_symtab_t structure 
 * Argument 2: the string to be resolved
 * Argument 3: a boolean specifying whether a new entry in the symbol table should be
 *             created, if no match is found (0: no, all other values: yes)
 * Returns:    the symbol ID, a non-negative integer, or -1, if no match is found 
 */
int template_symtab_get(template_symtab_t *symtab,const char *str,void **info,int create);
int template_symtab_apply(template_symtab_t *symtab,int (*f)(int sym, const char *str, void *info, void *data),void *data);
/*
 * Destroys a symbol table in caller-provided memory (i.e., the template_symtab_t structure
 * itself is not freed.
 * Argument 1: a pointer to the template_symtab_t structure
 */
void template_symtab_deinit(template_symtab_t *symtab);
/*
 * Frees a symbol table.
 * Argument 1: the pointer to the template_symtab_t structure
 */
void template_symtab_free(template_symtab_t *symtab);


#endif
